from rest_framework.permissions import BasePermission

from .verification import is_verified, not_have_active_verification
from .validators import validate_verification_type


class IsNotVerified(BasePermission):
    message = 'The user should be verified'

    def has_permission(self, request, view):
        user = request.user
        verification_type = validate_verification_type(request.data.get('verification_type', None))
        if user and user.is_authenticated:
            return is_verified(user, verification_type=verification_type)
        return False

    def has_object_permission(self, request, view, obj):
        user = request.user
        verification_type = validate_verification_type(request.data.get('verification_type', None))
        if user and user.is_authenticated:
            return is_verified(user, verification_type=verification_type)
        return False


class HasNotActiveVerification(BasePermission):
    message = 'The user has active verification.'

    def has_permission(self, request, view):
        user = request.user
        verification_type = validate_verification_type(request.data.get('verification_type', None))
        if user and user.is_authenticated:
            # This line may raise exception if user has active verification
            not_have_active_verification(user, verification_type=verification_type)
            return True
        return False

    def has_object_permission(self, request, view, obj):
        user = request.user
        verification_type = validate_verification_type(request.data.get('verification_type', None))
        if user and user.is_authenticated:
            # This line may raise exception if user has active verification
            not_have_active_verification(user, verification_type=verification_type)
            return True
        return False



from django.conf import settings
from django.db import models
from django.contrib.auth import get_user_model

User = get_user_model()


class BaseVerification(models.Model):
    user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_related",
        related_query_name="%(app_label)s_%(class)ss",
    )

    verification_type = models.CharField(
        max_length=128,
    )

    verification_code = None

    code_sent = models.BooleanField(
        default=False,
    )

    created = models.DateTimeField(
        auto_created=True,
        auto_now=True,
    )

    updated = models.DateTimeField(
        auto_created=True,
        auto_now=True,
    )

    class Meta:
        abstract = True
        ordering = ['created']


class Verification(BaseVerification):

    verification_code = models.CharField(
        max_length=32,
        unique=True,
    )

    class Meta:
        unique_together = ['user', 'verification_type']
        ordering = ['-created']


class VerificationRecord(BaseVerification):
    verification_code = models.CharField(
        max_length=32,
    )

    is_verified = models.BooleanField(
        default=False,
    )

    class Meta:
        ordering = ['-created']



from django.utils import timezone

default_config = {
    'VERIFICATIONS': [],
    'CONTAINS_NUMERIC': True,
    'CONTAINS_UPPER_ALPHABETIC': False,
    'CONTAINS_LOWER_ALPHABETIC': False,
    'CUSTOM_CODE_LETTERS': '',
    'CODE_LENGTH': 6,
    'LIFE_TIME_SECOND': 0,
    'LIFE_TIME_MINUTE': 3,
    'LIFE_TIME_HOUR': 0,
    'LIFE_TIME_DAY': 0,
    'LIFE_TIME_PENALTY_SECOND': 60,
}


CELERY_BEAT_SCHEDULE = {
    'archive-expired-verifications': {
        'task': 'verification.tasks.archive_expired_verifications',
        'schedule': timezone.timedelta(minutes=3),
    },
}

numeric = '0123456789'
lowercase_alphabetic = 'abcdefghijklmnopqrstuvwxyz'
uppercase_alphabetic = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'



from django.conf import settings
from django.utils import timezone
from .base import default_config, numeric, lowercase_alphabetic, uppercase_alphabetic

config = default_config
config.update(settings.VERIFICATION)

VERIFICATION_CODE_FIELD = 'verification_code'
VERIFICATIONS = config.get('VERIFICATIONS')
CODE_LENGTH = config.get('CODE_LENGTH')
LIFE_TIME_SECOND = config.get('LIFE_TIME_SECOND')
LIFE_TIME_MINUTE = config.get('LIFE_TIME_MINUTE')
LIFE_TIME_HOUR = config.get('LIFE_TIME_HOUR')
LIFE_TIME_DAY = config.get('LIFE_TIME_DAY')
LIFE_TIME_PENALTY_SECOND = config.get('LIFE_TIME_PENALTY_SECOND')

CODE_LIEF_TIME = timezone.timedelta(
    seconds=LIFE_TIME_SECOND + LIFE_TIME_PENALTY_SECOND,
    minutes=LIFE_TIME_MINUTE,
    hours=LIFE_TIME_HOUR,
    days=LIFE_TIME_DAY,
)

ALLOWED_CODE_LETTERS = ''
if config.get('CONTAINS_NUMERIC'):
    ALLOWED_CODE_LETTERS += numeric
if config.get('CONTAINS_UPPER_ALPHABETIC'):
    ALLOWED_CODE_LETTERS += uppercase_alphabetic
if config.get('CONTAINS_LOWER_ALPHABETIC'):
    ALLOWED_CODE_LETTERS += lowercase_alphabetic

if len(ALLOWED_CODE_LETTERS) == 0:
    raise Exception("No letters are allowed for code generation")


VERIFICATIONS_DICT = {}
VERIFICATIONS_TYPES = []
VERIFICATIONS_USER_MODEL_FIELDS = []
for verification in VERIFICATIONS:
    VERIFICATIONS_TYPES.append(verification.get('type'))
    VERIFICATIONS_USER_MODEL_FIELDS.append(verification.get('user_model_field'))
    VERIFICATIONS_DICT[verification.get('type')] = verification


def get_user_model_field(verification_type):
    return VERIFICATIONS_DICT.get(verification_type, None).get('user_model_field')




from django.utils.translation import ugettext_lazy as _
from rest_framework import status
from rest_framework.exceptions import APIException, ValidationError


# class UserIsVerified(Exception):
#     """The user already has been verified"""
#     pass


class UserIsVerified(APIException):
    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = _('The user has been already verified')
    default_code = 'invalid'


class VerificationTypeNotValid(APIException):
    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = _('Verification_field is not valid')
    default_code = 'invalid'


class UserHasActiveVerification(APIException):
    status_code = status.HTTP_403_FORBIDDEN
    default_detail = _('User has unexpired verification code')
    default_code = 'permission_denied'


class VerificationFailed(ValidationError):
    default_detail = _('Verification failed.')



from django.utils.translation import ugettext_lazy as _

from rest_framework import serializers
from rest_framework.exceptions import ValidationError, NotAuthenticated

from .models import (BaseVerification,
                     Verification, VerificationRecord)
from .validators import validate_verification_type

from .verification import generate_verification
from .conf import VERIFICATION_CODE_FIELD


class BaseVerificationSerializer(serializers.ModelSerializer):

    class Meta:
        model = BaseVerification
        fields = [
            'id', 'user', 'verification_type', 'verification_code', 'created', 'updated', VERIFICATION_CODE_FIELD
        ]

    def validate_verification_type(self, value):
        return validate_verification_type(value)

    def validate_user(self, value):
        request = self.context.get('request')
        user = request.user
        if not user:
            raise NotAuthenticated
        if user != value:
            raise ValidationError(_("The user should send request himself."))
        return value

    def validate(self, attrs):
        user = attrs
        is_verified(user, verification_type=self.verification_type)

    def create(self, validated_data):
        pass

    def update(self, instance, validated_data):
        pass


class VerificationSerializer(BaseVerificationSerializer):

    class Meta(BaseVerificationSerializer.Meta):
        model = Verification
        fields = [
            'id', 'user', 'verification_type', 'created', 'updated', VERIFICATION_CODE_FIELD
        ]

        extra_kwargs = {
            VERIFICATION_CODE_FIELD: {'write_only': True}
        }

    def create(self, validated_data):
        return generate_verification(validated_data['user'], validated_data['verification_type'])


# class VerificationSerializer(BaseVerificationSerializer):
#
#     class Meta(BaseVerificationSerializer.Meta):
#         model = Verification
#         fields = [
#             'id', 'user', 'verification_type', 'created', 'updated',
#         ]


class VerificationRecordSerializer(BaseVerificationSerializer):
    class Meta(BaseVerificationSerializer.Meta):
        model = VerificationRecord
        fields = [
            'id', 'user', 'verification_type', 'verification_code', 'created', 'updated', 'is_verified',
        ]




from celery import shared_task

from django.utils import timezone

from .models import Verification
from .conf import CODE_LIEF_TIME
from .verification import change_to_unverified_record


@shared_task()
def archive_expired_verifications():
    expired_verifications_qs = Verification.objects.filter(created__lt=timezone.now() - CODE_LIEF_TIME)
    for verification in expired_verifications_qs:
        change_to_unverified_record(verification=verification)



from rest_framework import generics, status
from rest_framework.response import Response
from rest_framework.settings import api_settings
from rest_framework.permissions import IsAuthenticated

from .serializers import VerificationSerializer
from .verification import verify
from .permissions import IsNotVerified, HasNotActiveVerification


class BaseGenerateVerificationAPIView(generics.GenericAPIView):
    send_code_function = None
    serializer_class = VerificationSerializer
    permission_classes = [IsAuthenticated, IsNotVerified, HasNotActiveVerification]

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        headers = self.get_success_headers(serializer.data)
        self.send_code_function(reuest=request, **serializer.validated_data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)

    def get_success_headers(self, data):
        try:
            return {'Location': str(data[api_settings.URL_FIELD_NAME])}
        except (TypeError, KeyError):
            return {}


class BaseVerifyVerificationAPIView(generics.GenericAPIView):
    serializer_class = VerificationSerializer
    permission_classes = [IsAuthenticated]

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        verified = verify(**serializer.validated_data)
        response_data = {}
        if not verified:
            response_data['verified'] = verified
            return Response(response_data, status=status.HTTP_200_OK)
        headers = self.get_success_headers(serializer.data)
        response_data.update(serializer.data)
        response_data['verified'] = verified
        return Response(response_data, status=status.HTTP_200_OK, headers=headers)

    def get_success_headers(self, data):
        try:
            return {'Location': str(data[api_settings.URL_FIELD_NAME])}
        except (TypeError, KeyError):
            return {}



# Generated by Django 3.1.1 on 2020-10-22 21:19

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='VerificationRecord',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('updated', models.DateTimeField(auto_created=True, auto_now=True)),
                ('created', models.DateTimeField(auto_created=True, auto_now=True)),
                ('verification_type', models.CharField(max_length=128)),
                ('code_sent', models.BooleanField(default=False)),
                ('verification_code', models.CharField(max_length=32)),
                ('is_verified', models.BooleanField(default=False)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='verification_verificationrecord_related', related_query_name='verification_verificationrecords', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ['-created'],
            },
        ),
        migrations.CreateModel(
            name='Verification',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('updated', models.DateTimeField(auto_created=True, auto_now=True)),
                ('created', models.DateTimeField(auto_created=True, auto_now=True)),
                ('verification_type', models.CharField(max_length=128)),
                ('code_sent', models.BooleanField(default=False)),
                ('verification_code', models.CharField(max_length=32, unique=True)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='verification_verification_related', related_query_name='verification_verifications', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ['-created'],
                'unique_together': {('user', 'verification_type')},
            },
        ),
    ]



from django.utils.crypto import get_random_string
from django.utils import timezone
from django.contrib.auth import get_user_model

from .conf import VERIFICATION_CODE_FIELD
from .models import Verification, VerificationRecord
from . import conf
from . import exceptions

User = get_user_model()


def generate_verification(user, verification_type) -> str:

    if is_verified(user, verification_type):
        raise exceptions.UserIsVerified

    try:
        expired_verification = not_have_active_verification(user, verification_type)

        change_to_unverified_record(expired_verification)

        verification_code = generate_unique_code(verification_type)
        new_verification = Verification.objects.create(user=user,
                                                       verification_type=verification_type,
                                                       verification_code=verification_code)
    except Verification.DoesNotExist:
        verification_code = generate_unique_code(verification_type)
        new_verification = Verification.objects.create(user=user,
                                                       verification_type=verification_type,
                                                       verification_code=verification_code)

    return new_verification


def verify(*args, **kwargs):
    user_id = kwargs.get('user')
    user = User.objcet.get(id=user_id)
    code = kwargs.get(VERIFICATION_CODE_FIELD)
    verification_type = kwargs.get('verification_type')
    return verify_code(user, verification_type, code)


def verify_code(user, verification_type, code) -> (bool, str):
    verified = False
    try:
        verification = Verification.objects.get(user=user,
                                                verification_type=verification_type)

        if verification.verification_code == code and timezone.now() < verification.created + conf.CODE_LIEF_TIME:
            change_to_verified_user(user, verification.verification_type)
            change_to_verified_record(verification)
            verified = True
        else:
            verified = False
    except Verification.DoesNotExist:
        verified = False

    return verified


def is_verified(user, verification_type):
    # last_verification_record = \
    #     VerificationRecord.objects.filter(user=user,
    #                                       verification_type=verification_type).order_by('created').last()
    # return last_verification_record.is_verified if last_verification_record is not None else False
    return getattr(user, conf.get_user_model_field(verification_type), False)


def not_have_active_verification(user, verification_type):
    """
    if user has active code exception will arise, else the expired verification will return
    :param user:
    :param verification_type:
    :return:
    """
    verification = Verification.objects.get(user=user,
                                            verification_type=verification_type)
    if timezone.now() < verification.created + conf.CODE_LIEF_TIME:
        raise exceptions.UserHasActiveVerification

    return verification


def change_to_verified_record(verification: Verification):
    VerificationRecord.objects.create(
        user=verification.user,
        verification_type=verification.verification_type,
        verification_code=verification.verification_code,
        is_verified=True,
    )

    verification.delete()


def change_to_unverified_record(verification: Verification):
    VerificationRecord.objects.create(
        user=verification.user,
        verification_type=verification.verification_type,
        verification_code=verification.verification_code,
        is_verified=False,
    )

    verification.delete()


def change_to_verified_user(user, verification_type):
    update_data = {conf.get_user_model_field(verification_type): True}
    User.objects.filter(pk=user.pk).update(**update_data)


def generate_unique_code(verification_type):
    code = get_random_string(length=conf.CODE_LENGTH, allowed_chars=conf.ALLOWED_CODE_LETTERS)
    qs_exists = Verification.objects.filter(verification_type=verification_type, verification_code=code).exists()
    if qs_exists:
        return generate_unique_code(verification_type)
    return code



from . import conf
from .exceptions import VerificationTypeNotValid


def validate_verification_type(value):
    if value not in conf.VERIFICATIONS_TYPES:
        raise VerificationTypeNotValid
    return value

